"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.printSchemaWithDirectives = void 0;
const graphql_1 = require("graphql");
function printSchemaWithDirectives(schema) {
    const typesMap = schema.getTypeMap();
    const queryType = schema.getQueryType();
    const mutationType = schema.getMutationType();
    const subscriptionType = schema.getSubscriptionType();
    const result = [
        `schema {
      ${queryType ? `query: ${queryType.toString()}` : ''}
      ${mutationType ? `mutation: ${mutationType.toString()}` : ''}
      ${subscriptionType ? `subscription: ${subscriptionType.toString()}` : ''}
    }`,
    ];
    for (const typeName in typesMap) {
        const type = typesMap[typeName];
        const isPredefinedScalar = graphql_1.isScalarType(type) && graphql_1.isSpecifiedScalarType(type);
        const isIntrospection = graphql_1.isIntrospectionType(type);
        if (isPredefinedScalar || isIntrospection) {
            continue;
        }
        // KAMIL: we might want to turn on descriptions in future
        result.push(graphql_1.print(correctType(typeName, typesMap).astNode));
    }
    const directives = schema.getDirectives();
    for (const directive of directives) {
        if (directive.astNode) {
            result.push(graphql_1.print(directive.astNode));
        }
    }
    return result.join('\n');
}
exports.printSchemaWithDirectives = printSchemaWithDirectives;
function extendDefinition(type) {
    const astNode = type.astNode;
    switch (astNode.kind) {
        case graphql_1.Kind.OBJECT_TYPE_DEFINITION:
            return {
                ...astNode,
                fields: astNode.fields.concat(type.extensionASTNodes.reduce((fields, node) => fields.concat(node.fields), [])),
            };
        case graphql_1.Kind.INPUT_OBJECT_TYPE_DEFINITION:
            return {
                ...astNode,
                fields: astNode.fields.concat(type.extensionASTNodes.reduce((fields, node) => fields.concat(node.fields), [])),
            };
        case graphql_1.Kind.ENUM_TYPE_DEFINITION:
            return {
                ...astNode,
                values: astNode.values.concat(type.extensionASTNodes.reduce((values, node) => values.concat(node.values), [])),
            };
        default:
            return astNode;
    }
}
function correctType(typeName, typesMap) {
    const type = typesMap[typeName];
    type.name = typeName.toString();
    if (type.astNode && type.extensionASTNodes) {
        type.astNode = type.extensionASTNodes ? extendDefinition(type) : type.astNode;
    }
    const doc = graphql_1.parse(graphql_1.printType(type));
    const fixedAstNode = doc.definitions[0];
    const originalAstNode = type === null || type === void 0 ? void 0 : type.astNode;
    if (originalAstNode) {
        // eslint-disable-next-line no-extra-semi
        ;
        fixedAstNode.directives = originalAstNode === null || originalAstNode === void 0 ? void 0 : originalAstNode.directives;
        if ('fields' in fixedAstNode && 'fields' in originalAstNode) {
            for (const fieldDefinitionNode of fixedAstNode.fields) {
                const originalFieldDefinitionNode = originalAstNode.fields.find((field) => field.name.value === fieldDefinitionNode.name.value);
                fieldDefinitionNode.directives = originalFieldDefinitionNode === null || originalFieldDefinitionNode === void 0 ? void 0 : originalFieldDefinitionNode.directives;
            }
        }
        if ('values' in fixedAstNode && 'values' in originalAstNode) {
            for (const enumValueDefinitionNode of fixedAstNode.values) {
                const originalEnumValueDefinitionNode = originalAstNode.values.find((enumValue) => enumValue.name.value === enumValueDefinitionNode.name.value);
                enumValueDefinitionNode.directives = originalEnumValueDefinitionNode === null || originalEnumValueDefinitionNode === void 0 ? void 0 : originalEnumValueDefinitionNode.directives;
            }
        }
    }
    type.astNode = fixedAstNode;
    return type;
}
//# sourceMappingURL=utils.js.map